@@ -18,13 +18,9 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
- *
  * 
  * Build with: make tcp6
  * 
- * This program has been tested to compile and run on: Debian GNU/Linux 6.0,
- * FreeBSD 8.2, NetBSD 5.1, OpenBSD 5.0, and Ubuntu 11.10.
- *
  * It requires that the libpcap library be installed on your system.
  *
  * Please send any bug reports to Fernando Gont <fgont@si6networks.com>
@@ -59,43 +55,55 @@
 #include "ipv6toolkit.h"
 #include "tcp6.h"
 #include "libipv6.h"
-#include <netinet/tcp.h>
 
 
 /* Function prototypes */
 void				init_packet_data(struct iface_data *);
-void				send_packet(struct iface_data *, const u_char *);
+int				is_valid_tcp_segment(struct iface_data *, const u_char *, struct pcap_pkthdr *);
+void				send_packet(struct iface_data *, const u_char *, struct pcap_pkthdr *);
 void				print_attack_info(struct iface_data *);
 void				usage(void);
 void				print_help(void);
 void				frag_and_send(struct iface_data *);
-
+unsigned int		queue_data(struct queue *, unsigned char *, unsigned int);
+unsigned int		dequeue_data(struct queue *, unsigned char *, unsigned int);
+unsigned int		queue_copy(struct queue *, unsigned char *, unsigned int, unsigned char *, unsigned int);
+unsigned int		queue_remove(struct queue *, unsigned char *, unsigned int);
+void				queue_purge( struct queue *);
+int					tcp_init(struct tcp *);
+int					tcp_open(struct iface_data *, struct tcp *, unsigned int);
+int					tcp_close(struct iface_data *, struct tcp *);
+int					tcp_send(struct iface_data *, struct tcp *, unsigned char *, unsigned int);
+int					tcp_receive(struct iface_data *, struct tcp *, unsigned char *, unsigned int);
+int					tcp_input(struct iface_data *, struct tcp *, const u_char *, struct pcap_pkthdr *, struct packet *);
+int					tcp_output(struct iface_data *, struct tcp *, struct packet *, struct timeval *);
+int					is_valid_tcp_segment(struct iface_data *, const u_char *, struct pcap_pkthdr *);
 
 /* Flags */
-unsigned char 		iface_f=0, floodt_f=0;
+unsigned char 		floodt_f=0;
 unsigned char 		listen_f=0, accepted_f=0, loop_f=0, sleep_f=0;
-unsigned char		srcprefix_f=0, hoplimit_f=0, rand_link_src_f=0, rand_src_f=0;
+unsigned char		hoplimit_f=0, rand_link_src_f=0, rand_src_f=0;
 unsigned char		floods_f=0, floodp_f=0, donesending_f=0, startclose_f=0;
 unsigned char		data_f=0, senddata_f=0, useaddrkey_f=0, window_f=0, winmodulate_f=0;
 
 /* Flags used for TCP (specifically) */ 
-unsigned int		srcport_f=0, dstport_f=0;
-unsigned int		tcpseq_f=0, tcpack_f=0, tcpurg_f=0, tcpflags_f=0, tcpwin_f=0;
-unsigned int		rhbytes_f=0, tcpflags_auto_f=0, tcpopen_f=0, tcpclose_f=0;
-unsigned int		tcpopen=0, tcpclose=0, win1_size=0, win2_size=0, pps_f=0, bps_f=0;
-unsigned int		ackdata_f=1, ackflags_f=1, window=0, time1_len=0, time2_len=0;
+unsigned char		srcport_f=0, dstport_f=0;
+unsigned char		tcpseq_f=0, tcpack_f=0, tcpurg_f=0, tcpflags_f=0, tcpwin_f=0;
+unsigned char		rhbytes_f=0, tcpflags_auto_f=0, tcpopen_f=0, tcpclose_f=0;
+unsigned char		pps_f=0, bps_f=0, debug_f=0, probe_f=0, retrans_f=0, rto_f=0;
+unsigned char		ackdata_f=1, ackflags_f=1;
+unsigned int		debug, tcpopen=0, tcpclose=0, win1_size=0, win2_size=0, window=0, time1_len=0, time2_len=0;
+
 u_int16_t			srcport, dstport, tcpurg, tcpwin, tcpwinm;
+unsigned int		retrans, rto;
 u_int32_t			tcpseq, tcpack;
-u_int8_t			tcpflags=0;
-struct tcphdr		*rhtcp;
+u_int8_t			tcpflags=0, pkt_tcp_flags;
+struct tcp_hdr		*rhtcp;
 unsigned int		rhbytes, currentsize, packetsize;
 
 
 /* Used for router discovery */
 struct iface_data	idata;
-struct prefix_entry	*prefix_ols[MAX_PREFIXES_ONLINK], *prefix_acs[MAX_PREFIXES_AUTO];
-struct prefix_entry	*prefix_local[MAX_LOCAL_ADDRESSES];
-
 
 /* Data structures for packets read from the wire */
 struct pcap_pkthdr		*pkthdr;
@@ -104,7 +112,7 @@ unsigned char			*pkt_end;
 struct ether_header		*pkt_ether;
 struct nd_neighbor_solicit	*pkt_ns;
 struct ip6_hdr			*pkt_ipv6;
-struct tcphdr			*pkt_tcp;
+struct tcp_hdr			*pkt_tcp;
 struct in6_addr			*pkt_ipv6addr;
 unsigned int			pktbytes;
 
@@ -122,7 +130,7 @@ char 				iface[IFACE_LENGTH];
 char				line[LINE_BUFFER_SIZE];
     
 struct ip6_hdr		*ipv6;
-struct tcphdr		*tcp;
+struct tcp_hdr		*tcp;
 
 struct ether_header	*ethernet;
 struct nd_opt_tlla	*tllaopt;
@@ -139,7 +147,7 @@ unsigned long		ul_res, ul_val, rate;
 unsigned int		i, j, startrand;
 unsigned int		skip;
 unsigned int		sources, nsources, ports, nports, nsleep;
-unsigned char		srcpreflen, randpreflen;
+unsigned char		randpreflen;
 
 u_int16_t			mask;
 u_int8_t			hoplimit;
@@ -171,10 +179,15 @@ struct filters		filters;
 int main(int argc, char **argv){
 	extern char		*optarg;	
 	char			*endptr; /* Used by strtoul() */
-	fd_set			sset, rset;
+	fd_set			sset, rset;	
+/*	fd_set			wset, eset; */
 	int				r, sel;
 	struct timeval	timeout, stimeout, curtime, lastprobe, wmtimeout;
-	unsigned long	pktinterval;
+	/*struct tcp		tcb; */
+	/* unsigned char	end_f=0, error_f; */
+	unsigned char		end_f=0;
+	unsigned long	pktinterval=0; /*Add  datasent=0*/
+	unsigned int	retr=0;
 
 	static struct option longopts[] = {
 		{"interface", required_argument, 0, 'i'},
@@ -218,11 +231,14 @@ int main(int argc, char **argv){
 		{"rate-limit", required_argument, 0, 'r'},
 		{"sleep", required_argument, 0, 'z'},
 		{"listen", no_argument, 0, 'L'},
+		{"probe", no_argument, 0, 'p'},
+		{"retrans", required_argument, 0, 'x'},
 		{"verbose", no_argument, 0, 'v'},
+		{"debug", required_argument, 0, 'Y'},
 		{"help", no_argument, 0, 'h'}
 	};
 
-	char shortopts[]= "i:s:d:A:c:C:Z:u:U:H:y:S:D:P:o:a:X:q:Q:V:w:W:M:Nnj:k:J:K:b:g:B:G:F:T:fRlr:z:Lvh";
+	char shortopts[]= "i:s:d:A:c:C:Z:u:U:H:y:S:D:P:o:a:X:q:Q:V:w:W:M:Nnj:k:J:K:b:g:B:G:F:T:fRlr:z:Lpx:vyY:h";
 
 	char option;
 
@@ -242,15 +258,20 @@ int main(int argc, char **argv){
 		exit(EXIT_FAILURE);
 	}
 
-	init_iface_data(&idata);
+	if(init_iface_data(&idata) == FAILURE){
+		puts("Error initializing internal data structure");
+		exit(EXIT_FAILURE);
+	}
 
 	while((r=getopt_long(argc, argv, shortopts, longopts, NULL)) != -1) {
 		option= r;
 
-		switch(option) {
+		switch(option){
 			case 'i':  /* Interface */
 				strncpy(idata.iface, optarg, IFACE_LENGTH-1);
-				iface_f=1;
+				idata.iface[IFACE_LENGTH-1]=0;
+				idata.ifindex= if_nametoindex(idata.iface);
+				idata.iface_f=TRUE;
 				break;
 
 			case 's':	/* IPv6 Source Address */
@@ -272,18 +293,18 @@ int main(int argc, char **argv){
 				idata.srcaddr_f = 1;
 		
 				if((charptr = strtok_r(NULL, " ", &lasts)) != NULL){
-					srcpreflen = atoi(charptr);
+					idata.srcpreflen = atoi(charptr);
 		
-					if(srcpreflen>128){
+					if(idata.srcpreflen>128){
 						puts("Prefix length error in IPv6 Source Address");
 						exit(EXIT_FAILURE);
 					}
 
-					if(srcpreflen == 64)
+					if(idata.srcpreflen == 64)
 						useaddrkey_f= 1;
 
-					sanitize_ipv6_prefix(&(idata.srcaddr), srcpreflen);
-					srcprefix_f=1;
+					sanitize_ipv6_prefix(&(idata.srcaddr), idata.srcpreflen);
+					idata.srcprefix_f=1;
 				}
 
 				break;
@@ -312,6 +333,9 @@ int main(int argc, char **argv){
 				else if(strncmp(optarg, "abort", MAX_CMDLINE_OPT_LEN) == 0){
 					tcpopen= OPEN_ABORT;
 				}
+				else if(strncmp(optarg, "active", MAX_CMDLINE_OPT_LEN) == 0){
+					tcpopen= OPEN_ACTIVE;
+				}
 				else{
 					puts("Error: Unknown open mode in '-c' option");
 					exit(EXIT_FAILURE);
@@ -356,10 +380,10 @@ int main(int argc, char **argv){
 			case 'Z': /* Data */
 				datalen= Strnlen(optarg, MAX_CMDLINE_OPT_LEN);
 
-				if(datalen > DATA_BUFFER_LEN)
-					datalen= DATA_BUFFER_LEN;
+				if(datalen >= DATA_BUFFER_LEN)
+					datalen= DATA_BUFFER_LEN-1;
 
-				strncpy(data, optarg, DATA_BUFFER_LEN);
+				strncpy(data, optarg, DATA_BUFFER_LEN-1);
 				data_f=1;
 				break;
 
@@ -930,7 +954,31 @@ int main(int argc, char **argv){
 			case 'v':	/* Be verbose */
 				(idata.verbose_f)++;
 				break;
-		
+
+			case 'p':	/* Probe mode */
+				probe_f=1;
+				break;
+
+			case 'x':	/* Number of retrnasmissions */
+				retrans= atoi(optarg);
+				retrans_f=1;
+				break;
+
+			case 'Y':
+				if(strncmp(optarg, "dump", MAX_CMDLINE_OPT_LEN) == 0){
+					debug= DEBUG_DUMP;
+				}
+				else if(strncmp(optarg, "script", MAX_CMDLINE_OPT_LEN) == 0){
+					debug= DEBUG_SCRIPT;
+				}
+				else{
+					puts("Error: Unknown open mode in '-Y' option");
+					exit(EXIT_FAILURE);
+				}
+
+				debug_f=1;
+				break;
+
 			case 'h':	/* Help */
 				print_help();
 				exit(EXIT_FAILURE);
@@ -949,48 +997,17 @@ int main(int argc, char **argv){
 		exit(EXIT_FAILURE);
 	}
 
-	if(!iface_f){
-		puts("Must specify the network interface with the -i option");
-		exit(EXIT_FAILURE);
-	}
-
-	if( (idata.pfd= pcap_open_live(idata.iface, PCAP_SNAP_LEN, PCAP_PROMISC, PCAP_TIMEOUT, errbuf)) == NULL){
-		printf("pcap_open_live(): %s\n", errbuf);
-		exit(EXIT_FAILURE);
-	}
-
-	release_privileges();
-
-	if( (idata.type = pcap_datalink(idata.pfd)) == DLT_EN10MB){
-		idata.linkhsize= ETH_HLEN;
-		idata.mtu= ETH_DATA_LEN;
-	}
-	else if( idata.type == DLT_RAW){
-		idata.linkhsize=0;
-		idata.mtu= MIN_IPV6_MTU;
-		idata.flags= IFACE_TUNNEL;
-	}
-	else if(idata.type == DLT_NULL){
-		idata.linkhsize=4;
-		idata.mtu= MIN_IPV6_MTU;
-		idata.flags= IFACE_TUNNEL;
-	}
-	else{
-		printf("Error: Interface %s is not an Ethernet or tunnel interface", iface);
-		exit(EXIT_FAILURE);
-	}
-
 	srandom(time(NULL));
 
 	/*
 	  If the flood option ("-F") has been specified, but no prefix has been specified,
-	  select the random Source Addresses from the link-local unicast prefix (fe80::/64).
+	  assume a /64 prefix.
 	*/
-	if(floods_f && !srcprefix_f){
-		srcpreflen=0;
+	if(floods_f && !idata.srcprefix_f){
+		idata.srcpreflen=64;
 	}
 
-	if(srcprefix_f && !floods_f && loop_f){
+	if(idata.srcprefix_f && !floods_f && loop_f){
 		floods_f=1;
 		nsources= 1;
 	}
@@ -1005,31 +1022,28 @@ int main(int argc, char **argv){
 		exit(EXIT_FAILURE);
 	}
 
-	if(get_if_addrs(&idata) == -1){
-		puts("Error obtaining local addresses");
+	if(!idata.iface_f){
+		if(idata.dstaddr_f && IN6_IS_ADDR_LINKLOCAL(&(idata.dstaddr))){
+			puts("Must specify a network interface for link-local destinations");
+			exit(EXIT_FAILURE);
+		}
+		else if(listen_f){
+			puts("Must specify a network interface when employing the 'listenging' mode");
+			exit(EXIT_FAILURE);
+		}
+	}
+
+	if(load_dst_and_pcap(&idata, (idata.dstaddr_f?LOAD_SRC_NXT_HOP:LOAD_PCAP_ONLY)) == FAILURE){
+		puts("Error while learning Souce Address and Next Hop");
 		exit(EXIT_FAILURE);
 	}
 
-	if(!idata.ether_flag){
-		randomize_ether_addr(&idata.ether);
-		idata.ether_flag=1;
-	}
-
-	if(!(idata.hsrcaddr_f)){
-		if(idata.ether_flag)
-			idata.hsrcaddr=idata.ether;
-		else
-			randomize_ether_addr(&(idata.hsrcaddr));
-	}
-
-	if(!idata.ip6_local_flag){
-		ether_to_ipv6_linklocal(&idata.ether, &idata.ip6_local);
-	}
+	release_privileges();
 
 	if(data_f){
 		data[datalen]=0;
 
-		if(!string_escapes(data, &datalen)){
+		if(!string_escapes(data, &datalen, DATA_BUFFER_LEN-1)){
 			puts("Error in data string option ('-Z')");
 			exit(EXIT_FAILURE);
 		}
@@ -1037,19 +1051,6 @@ int main(int argc, char **argv){
 		data[datalen]=0;
 	}
 
-	if(sel_next_hop(&idata) == -1)
-		exit(EXIT_FAILURE);
-
-	if(srcprefix_f && !floods_f){
-		randprefix=idata.srcaddr;
-		randpreflen=srcpreflen;
-		randomize_ipv6_addr(&(idata.srcaddr), &randprefix, randpreflen);
-		idata.srcaddr_f=1;
-	}
-	else if(!(idata.srcaddr_f)){
-		idata.srcaddr= *src_addr_sel(&idata, &(idata.dstaddr));
-	}
-
 	if(!floods_f)
 		nsources=1;
 
@@ -1077,7 +1078,7 @@ int main(int argc, char **argv){
 	}
 
 	if(bps_f){
-		packetsize= MIN_IPV6_HLEN +  sizeof(struct tcphdr) + rhbytes;
+		packetsize= MIN_IPV6_HLEN +  sizeof(struct tcp_hdr) + rhbytes;
 
 		for(i=0; i < ndstopthdr; i++)
 			packetsize+= dstopthdrlen[i];
@@ -1191,6 +1192,121 @@ int main(int argc, char **argv){
 		tcpwinm= win1_size;
 	}
     
+
+	if(probe_f){
+		end_f=0;
+
+		if(!dstport_f)
+			dstport= 80;
+
+		if(!srcport_f)
+			srcport= 50000 + random() % 15000; /* We select ports from the "high ports" range */
+
+		if(!rto_f)
+			rto=1;
+
+		if(!retrans_f)
+			retrans=0;
+
+		retr=0;
+		retrans++;
+
+		FD_ZERO(&sset);
+		FD_SET(idata.fd, &sset);
+
+		lastprobe.tv_sec= 0;	
+		lastprobe.tv_usec=0;
+
+		while(!end_f){
+			if(gettimeofday(&curtime, NULL) == -1){
+				if(idata.verbose_f)
+					perror("tcp6");
+
+				exit(EXIT_FAILURE);
+			}			
+
+			if(is_time_elapsed(&curtime, &lastprobe, 1) && retr < retrans){
+				retr++;
+				lastprobe= curtime;
+				send_packet(&idata, NULL, NULL);
+			}
+
+			if(is_time_elapsed(&curtime, &lastprobe, rto) && retr >= retrans){
+				end_f=1;
+				break;
+			}
+
+			rset= sset;
+			timeout.tv_usec=0;
+			timeout.tv_sec= (rto < 1)?rto:1;
+
+			if((sel=select(idata.fd+1, &rset, NULL, NULL, &timeout)) == -1){
+				if(errno == EINTR){
+					continue;
+				}
+				else{
+					puts("Error in select()");
+					exit(EXIT_FAILURE);
+				}
+			}
+
+			if(sel){
+				if(FD_ISSET(idata.fd, &rset)){
+					/* Read a packet */
+					if((r=pcap_next_ex(idata.pfd, &pkthdr, &pktdata)) == -1){
+						printf("pcap_next_ex(): %s", pcap_geterr(idata.pfd));
+						exit(EXIT_FAILURE);
+					}
+					else if(r == 0){
+						continue; /* Should never happen */
+					}
+
+					pkt_ether = (struct ether_header *) pktdata;
+					pkt_ipv6 = (struct ip6_hdr *)((char *) pkt_ether + idata.linkhsize);
+					pkt_tcp= (struct tcp_hdr *) ( (char *) pkt_ipv6 + MIN_IPV6_HLEN);
+					pkt_ns= (struct nd_neighbor_solicit *) ( (char *) pkt_ipv6 + MIN_IPV6_HLEN);
+					pkt_end = (unsigned char *) pktdata + pkthdr->caplen;
+					pkt_tcp_flags= pkt_tcp->th_flags;
+
+					/* Check that we are able to look into the IPv6 header */
+					if( (pkt_end -  pktdata) < (idata.linkhsize + MIN_IPV6_HLEN))
+						continue;
+
+					if(is_eq_in6_addr(&(pkt_ipv6->ip6_src), &(idata.srcaddr))){
+						continue;
+					}
+
+					if(!is_eq_in6_addr(&(pkt_ipv6->ip6_dst), &(idata.srcaddr))){
+						continue;
+					}
+
+					if(pkt_tcp->th_sport != htons(dstport)){
+						continue;
+					}
+
+					if(pkt_tcp->th_dport != htons(srcport)){
+						continue;
+					}
+
+					/* The TCP checksum must be valid */
+					if(in_chksum(pkt_ipv6, pkt_tcp, pkt_end-((unsigned char *)pkt_tcp), IPPROTO_TCP) != 0)
+						continue;
+
+					printf("PROBE:RESPONSE:%s%s%s%s%s%s\n", ((pkt_tcp_flags & TH_FIN)?"F":""), \
+						((pkt_tcp_flags & TH_SYN)?"S":""), \
+						((pkt_tcp_flags & TH_RST)?"R":""), ((pkt_tcp_flags & TH_PUSH)?"P":""),\
+						((pkt_tcp_flags & TH_ACK)?"A":""), ((pkt_tcp_flags & TH_URG)?"U":""));
+
+					exit(0);
+				}
+			}
+		}
+
+		puts("PROBE:TIMEOUT:");
+		exit(0);
+	}
+
+
 	/* Fire a TCP segment if an IPv6 Destination Address was specified */
 	if(!listen_f && idata.dstaddr_f){
 		if(loop_f){
@@ -1199,10 +1315,10 @@ int main(int argc, char **argv){
 											((nsleep>1)?"s":""));
 		}
 
-		while(!donesending_f){
-				send_packet(&idata, NULL);
+		do{
+				send_packet(&idata, NULL, NULL);
 
-				if((sel=select(0, NULL, NULL, NULL, &timeout)) == -1){
+				if(loop_f && (sel=select(0, NULL, NULL, NULL, &timeout)) == -1){
 					if(errno == EINTR){
 						continue;
 					}
@@ -1211,7 +1327,7 @@ int main(int argc, char **argv){
 						exit(EXIT_FAILURE);
 					}
 				}
-		}
+		}while(loop_f);
 
 		if(idata.verbose_f)    
 			puts("Initial attack packet(s) sent successfully.");
@@ -1284,7 +1400,7 @@ int main(int argc, char **argv){
 
 				pkt_ether = (struct ether_header *) pktdata;
 				pkt_ipv6 = (struct ip6_hdr *)((char *) pkt_ether + idata.linkhsize);
-				pkt_tcp= (struct tcphdr *) ( (char *) pkt_ipv6 + MIN_IPV6_HLEN);
+				pkt_tcp= (struct tcp_hdr *) ( (char *) pkt_ipv6 + MIN_IPV6_HLEN);
 				pkt_ns= (struct nd_neighbor_solicit *) ( (char *) pkt_ipv6 + MIN_IPV6_HLEN);
 				pkt_end = (unsigned char *) pktdata + pkthdr->caplen;
 
@@ -1366,7 +1482,7 @@ int main(int argc, char **argv){
 
 				if(pkt_ipv6->ip6_nxt == IPPROTO_TCP){
 					/* Check that we are able to look into the TCP header */
-					if( (pkt_end -  pktdata) < (idata.linkhsize + MIN_IPV6_HLEN + sizeof(struct tcphdr))){
+					if( (pkt_end -  pktdata) < (idata.linkhsize + MIN_IPV6_HLEN + sizeof(struct tcp_hdr))){
 						continue;
 					}
 
@@ -1400,6 +1516,10 @@ int main(int argc, char **argv){
 							}
 						}
 
+						/* The TCP checksum must be valid */
+						if(in_chksum(pkt_ipv6, pkt_tcp, pkt_end-((unsigned char *)pkt_tcp), IPPROTO_TCP) != 0)
+							continue;
+
 						if(pkt_tcp->th_sport != htons(dstport)){
 							continue;
 						}
@@ -1410,7 +1530,7 @@ int main(int argc, char **argv){
 					}
 
 					/* Send a TCP segment */
-					send_packet(&idata, pktdata);
+					send_packet(&idata, pktdata, pkthdr);
 				}
 				else if(pkt_ipv6->ip6_nxt == IPPROTO_ICMPV6){
 
@@ -1429,7 +1549,7 @@ int main(int argc, char **argv){
 							}
 
 							/* Check that the target address belongs to the prefix from which we are sending packets */
-							if(!match_ipv6(&(idata.srcaddr), &srcpreflen, 1, &(pkt_ns->nd_ns_target))){
+							if(!match_ipv6(&(idata.srcaddr), &idata.srcpreflen, 1, &(pkt_ns->nd_ns_target))){
 								continue;
 							}
 						}
@@ -1448,7 +1568,7 @@ int main(int argc, char **argv){
 			}
 			if((!sel || is_time_elapsed(&curtime, &lastprobe, pktinterval)) && !donesending_f){
 				lastprobe= curtime;
-				send_packet(&idata, NULL);
+				send_packet(&idata, NULL, NULL);
 			}
 		}
     
@@ -1479,7 +1599,7 @@ void init_packet_data(struct iface_data *idata){
 	if(idata->flags != IFACE_TUNNEL && idata->flags != IFACE_LOOPBACK){
 		ethernet->src = idata->hsrcaddr;
 		ethernet->dst = idata->hdstaddr;
-		ethernet->ether_type = htons(0x86dd);
+		ethernet->ether_type = htons(ETHERTYPE_IPV6);
 	}
 
 	ipv6->ip6_flow=0;
@@ -1577,7 +1697,7 @@ void init_packet_data(struct iface_data *idata){
  *
  * Initialize the remaining fields of the TCP segment, and send the attack packet(s).
  */
-void send_packet(struct iface_data *idata, const u_char *pktdata){
+void send_packet(struct iface_data *idata, const u_char *pktdata, struct pcap_pkthdr *pkthdr){
 	static unsigned int	sources=0, ports=0;	
 	ptr=startofprefixes;
 
@@ -1587,8 +1707,8 @@ void send_packet(struct iface_data *idata, const u_char *pktdata){
 	if(pktdata != NULL){   /* Sending a TCP segment in response to a received packet */
 		pkt_ether = (struct ether_header *) pktdata;
 		pkt_ipv6 = (struct ip6_hdr *)((char *) pkt_ether + idata->linkhsize);
-		pkt_tcp= (struct tcphdr *)( (char *) pkt_ipv6 + sizeof(struct ip6_hdr));
-		pkt_end = (unsigned char *) pktdata + pkthdr->len;
+		pkt_tcp= (struct tcp_hdr *)( (char *) pkt_ipv6 + sizeof(struct ip6_hdr));
+		pkt_end = (unsigned char *) pktdata + pkthdr->caplen;
 
 		/* The packet length is the minimum of what we capured, and what is specified in the
 		   IPv6 Total Lenght field
@@ -1630,7 +1750,7 @@ void send_packet(struct iface_data *idata, const u_char *pktdata){
 		}
 
 
-		if( (ptr+sizeof(struct tcphdr)) > (v6buffer+ idata->max_packet_size)){
+		if( (ptr+sizeof(struct tcp_hdr)) > (v6buffer+ idata->max_packet_size)){
 			puts("Packet Too Large while inserting TCP header");
 			exit(EXIT_FAILURE);
 		}
@@ -1639,8 +1759,8 @@ void send_packet(struct iface_data *idata, const u_char *pktdata){
 		if((tcpflags_auto_f || tcpopen_f || tcpclose_f) && pkt_tcp->th_flags & TH_RST)
 			return;
 
-		tcp = (struct tcphdr *) ptr;
-		bzero(tcp, sizeof(struct tcphdr));
+		tcp = (struct tcp_hdr *) ptr;
+		bzero(tcp, sizeof(struct tcp_hdr));
 
 		tcp->th_sport= pkt_tcp->th_dport;
 		tcp->th_dport= pkt_tcp->th_sport;
@@ -1877,7 +1997,7 @@ void send_packet(struct iface_data *idata, const u_char *pktdata){
 		tcp->th_urp= htons(tcpurg);
 
 		/* Current version of tcp6 does not support sending TCP options */
-		tcp->th_off= sizeof(struct tcphdr) >> 2;
+		tcp->th_off= sizeof(struct tcp_hdr) >> 2;
 		ptr+= tcp->th_off << 2;
 
 		if(rhbytes){
@@ -1906,7 +2026,7 @@ void send_packet(struct iface_data *idata, const u_char *pktdata){
 
 		if(senddata_f){
 			tcp->th_seq= htonl( ntohl(tcp->th_seq) + ptr-((unsigned char *)tcp + (tcp->th_off << 2)));
-			ptr= (unsigned char *)tcp + sizeof(struct tcphdr);
+			ptr= (unsigned char *)tcp + sizeof(struct tcp_hdr);
 
 			if((ptr+ datalen) > (v6buffer + idata->max_packet_size)){
 				if(idata->verbose_f)
@@ -1934,7 +2054,7 @@ void send_packet(struct iface_data *idata, const u_char *pktdata){
 
 		if(startclose_f){
 			tcp->th_seq= htonl( ntohl(tcp->th_seq) + ptr-((unsigned char *)tcp + (tcp->th_off << 2)));
-			ptr= (unsigned char *) tcp + sizeof(struct tcphdr);
+			ptr= (unsigned char *) tcp + sizeof(struct tcp_hdr);
 
 			if(tcpclose == CLOSE_ABORT){
 				tcp->th_flags= TH_ACK | TH_RST;
@@ -1966,13 +2086,13 @@ void send_packet(struct iface_data *idata, const u_char *pktdata){
 			}
 		}
 
-		if( (ptr+sizeof(struct tcphdr)) > (v6buffer + idata->max_packet_size)){
+		if( (ptr+sizeof(struct tcp_hdr)) > (v6buffer + idata->max_packet_size)){
 			puts("Packet Too Large while inserting TCP header");
 			exit(EXIT_FAILURE);
 		}
 
-		tcp= (struct tcphdr *) ptr;
-		bzero(ptr, sizeof(struct tcphdr));
+		tcp= (struct tcp_hdr *) ptr;
+		bzero(ptr, sizeof(struct tcp_hdr));
 		tcp->th_sport= htons(srcport);
 		tcp->th_dport= htons(dstport);
 		tcp->th_seq = htonl(tcpseq);
@@ -1991,7 +2111,7 @@ void send_packet(struct iface_data *idata, const u_char *pktdata){
 
 		tcp->th_urp= htons(tcpurg);
 		tcp->th_win= htons(tcpwin);
-		tcp->th_off= sizeof(struct tcphdr) >> 2;
+		tcp->th_off= sizeof(struct tcp_hdr) >> 2;
 
 		ptr += tcp->th_off << 2;
 
@@ -2019,7 +2139,7 @@ void send_packet(struct iface_data *idata, const u_char *pktdata){
 			   "srcaddr" and srcpreflen.
 			 */  
 
-			randomize_ipv6_addr( &(ipv6->ip6_src), &(idata->srcaddr), srcpreflen);
+			randomize_ipv6_addr( &(ipv6->ip6_src), &(idata->srcaddr), idata->srcpreflen);
 
 			/*
 			   If we need to respond to incomming packets, we set the Interface ID such that we can
@@ -2204,9 +2324,6 @@ void print_help(void){
 }
 
 
-
-
-
 /*
  * Function: print_attack_info()
  *
@@ -2269,8 +2386,8 @@ void print_attack_info(struct iface_data *idata){
 		}
 	}
 	else{
-		printf("IPv6 Source Address: randomized, from the %s/%u prefix%s\n", psrcaddr, srcpreflen, \
-    									(!srcprefix_f)?" (default)":"");
+		printf("IPv6 Source Address: randomized, from the %s/%u prefix%s\n", psrcaddr, idata->srcpreflen, \
+    									(!idata->srcprefix_f)?" (default)":"");
 	}
 
 	if(idata->dstaddr_f){
@@ -2388,29 +2505,443 @@ void print_attack_info(struct iface_data *idata){
 
 
 /*
- * Function: init_iface_data()
+ * Function: queue_data()
  *
- * Initializes the contents of "iface_data" structure
+ * Puts data into a queue
  */
 
-int init_iface_data(struct iface_data *idata){
-	bzero(idata, sizeof(struct iface_data));
-	idata->local_retrans = 0;
-	idata->local_timeout = 1;
+unsigned int queue_data(struct queue *q, unsigned char *data, unsigned int nbytes){
+	unsigned int	fbytes, nleft;
 
-	idata->ip6_global.prefix= prefix_local;
-	idata->ip6_global.nprefix=0;
-	idata->ip6_global.maxprefix= MAX_LOCAL_ADDRESSES;
+	/*
+	   We have to scenarios: in >= out and in < out
+	   In the first scenario, the circular buffer may be "split in two". In the second one,
+	   all available space is clustered together.
+	 */
+	if(q->in >= q->out){
+		fbytes= (q->data+ q->size) - q->in -1;
+		fbytes= fbytes+ (q->out - q->data);
 
-	idata->prefix_ol.prefix= prefix_ols;
-	idata->prefix_ol.nprefix= 0;
-	idata->prefix_ol.maxprefix= MAX_PREFIXES_ONLINK;
+		if(nbytes > fbytes)
+			nbytes= fbytes;
+		
+		/* There is enough space available on the right side of the buffer */
+		if( (q->data + q->size - q->in) >= nbytes){
+			memcpy(q->in, data, nbytes);
 
-	idata->prefix_ac.prefix= prefix_acs;
-	idata->prefix_ac.nprefix= 0;
-	idata->prefix_ac.maxprefix= MAX_PREFIXES_AUTO;
+			q->in= q->in + nbytes;
+			
+			if(q->in == (q->data + q->size))
+				q->in= q->data;
 
-	return 0;
+			return(nbytes);
+		}
+		else{
+			nleft= nbytes;
+			memcpy(q->in, data, (q->data + q->size - q->in));
+
+			nleft= nleft - (q->data + q->size - q->in);
+			q->in= q->data;
+
+			memcpy(q->in, data, nleft);
+			return(nbytes);
+		}
+	}
+	else{
+		fbytes= q->out - q->in - 1;
+
+		if(nbytes > fbytes)
+			nbytes= fbytes;
+
+		memcpy(q->in, data, nbytes);
+		q->in= q->in + nbytes;
+
+		if(q->in == (q->data + q->size))
+			q->in= q->data;
+
+		return(nbytes);
+	}
+
+	/* Should never reach here, but avoid compiler warnings */
+	return(0);
 }
 
 
+/*
+ * Function: dequeue_data()
+ *
+ * Reads data from a queue
+ */
+
+unsigned int dequeue_data(struct queue *q, unsigned char *data, unsigned int nbytes){
+	unsigned int	dbytes, nleft;
+
+	/*
+	   We have to scenarios: out > in and out <= in
+	   In the first scenario, the circular buffer may be "split in two". In the second one,
+	   all available data are clustered together.
+	 */
+	if(q->out > q->in){
+		dbytes= (q->data+ q->size) - q->out;
+		dbytes= dbytes+ (q->in - q->out);
+
+		if(nbytes > dbytes)
+			nbytes= dbytes;
+		
+		/* There is enough data available on the right side of the buffer */
+		if( (q->data + q->size - q->out) >= nbytes){
+			memcpy(data, q->out, nbytes);
+
+			q->out= q->out + nbytes;
+			
+			if(q->out == (q->data + q->size))
+				q->out= q->data;
+
+			return(nbytes);
+		}
+		else{
+			/* Data are split in two parts */
+			nleft= nbytes;
+			memcpy(data, q->out, (q->data + q->size - q->out));
+			data= data+ (q->data + q->size - q->out);
+
+			nleft= nleft - (q->data + q->size - q->out);
+			q->out= q->data;
+
+			memcpy(data, q->out, nleft);
+			q->out= q->out + nleft;
+
+			return(nbytes);
+		}
+	}
+	else{
+		dbytes= q->in - q->out;
+
+		if(nbytes > dbytes)
+			nbytes= dbytes;
+
+		memcpy(data, q->out, nbytes);
+		q->out= q->out + nbytes;
+
+		if(q->out == (q->data + q->size))
+			q->out= q->data;
+
+		return(nbytes);
+	}
+
+	/* Should never reach here, but avoid compiler warnings */
+	return(0);
+}
+
+
+
+/*
+ * Function: queue_copy()
+ *
+ * Copies data from queue, without removing it
+ */
+
+unsigned int queue_copy(struct queue *q, unsigned char *org, unsigned int offset, unsigned char *data, unsigned int nbytes){
+	unsigned int	dbytes, nleft;
+
+	if(org+offset >= (q->data + q->size)){
+		org= q->data + offset - (q->data + q->size - org);
+	}
+
+	/* | in    out
+	   We have to scenarios: out > in and out <= in
+	   In the first scenario, the circular buffer may be "split in two". In the second one,
+	   all available data are clustered together.
+	 */
+	if(org > q->in){
+		dbytes= (q->data+ q->size) - org;
+		dbytes= dbytes+ (q->in - org);
+
+		if(nbytes > dbytes)
+			nbytes= dbytes;
+		
+		/* There is enough data available on the right side of the buffer */
+		if( (q->data + q->size - org) >= nbytes){
+			memcpy(data, org, nbytes);
+			return(nbytes);
+		}
+		else{
+			/* Data are split in two parts */
+			nleft= nbytes;
+			memcpy(data, org, (q->data + q->size - org));
+			data= data + (q->data + q->size - org);
+
+			nleft= nleft - (q->data + q->size - org);
+			org= q->data;
+
+			memcpy(data, org, nleft);
+			return(nbytes);
+		}
+	}
+	else{
+		dbytes= q->in - org;
+
+		if(nbytes > dbytes)
+			nbytes= dbytes;
+
+		memcpy(data, org, nbytes);
+		return(nbytes);
+	}
+
+	/* Should never reach here, buts avoid compiler warnings */
+	return(0);
+}
+
+
+/*
+ * Function: queue_remove()
+ *
+ * Discards data from queue
+ * Note: This function is employed to discard data from the TCP send buffer when they are ACKed
+ * by the remote TCP endpoint.
+ */
+
+unsigned int queue_remove(struct queue *q, unsigned char *data, unsigned int nbytes){
+	unsigned int	dbytes, nleft;
+
+	/*
+	   We have to scenarios: out > in and out <= in
+	   In the first scenario, the circular buffer may be "split in two". In the second one,
+	   all available data are clustered together.
+	 */
+	if(q->out > q->in){
+		dbytes= (q->data+ q->size) - q->out;
+		dbytes= dbytes+ (q->in - q->out);
+
+		if(nbytes > dbytes)
+			nbytes= dbytes;
+		
+		/* There is enough data available on the right side of the buffer */
+		if( (q->data + q->size - q->out) >= nbytes){
+			q->out= q->out + nbytes;
+			
+			if(q->out == (q->data + q->size))
+				q->out= q->data;
+
+			return(nbytes);
+		}
+		else{
+			/* Data are split in two parts */
+			nleft= nbytes;
+			data= data+ (q->data + q->size - q->out);
+			nleft= nleft - (q->data + q->size - q->out);
+			q->out= q->data;
+			q->out= q->out + nleft;
+			return(nbytes);
+		}
+	}
+	else{
+		dbytes= q->in - q->out;
+
+		if(nbytes > dbytes)
+			nbytes= dbytes;
+
+		memcpy(data, q->out, nbytes);
+		q->out= q->out + nbytes;
+
+		if(q->out == (q->data + q->size))
+			q->out= q->data;
+
+		return(nbytes);
+	}
+
+	/* Should never reach here, but avoid compiler warnings */
+	return(0);
+}
+
+
+
+/*
+ * Function: tcp_init()
+ *
+ * Initilizes a TCP structure
+ */
+
+int tcp_init(struct tcp *tcp){
+	memset(&(tcp->srcaddr), 0, sizeof(struct in6_addr));
+	memset(&(tcp->dstaddr), 0, sizeof(struct in6_addr));
+	tcp->srcport= 0;
+	tcp->dstport= 0;
+
+	tcp->in.in = tcp->in.data;
+	tcp->in.out= tcp->in.data;
+	tcp->in.size= sizeof(tcp->in.data);
+
+	tcp->rcv_nxt= 0;
+	tcp->rcv_nxtwnd= 0;
+
+	tcp->out.in= tcp->out.data;
+	tcp->out.out= tcp->out.data;
+	tcp->out.size= sizeof(tcp->out.data);
+
+	tcp->out_una= tcp->out.data;
+	tcp->out_nxt= tcp->out.data;
+
+	tcp->snd_una=0;
+	tcp->snd_nxtwnd=0;
+	
+	memset(&(tcp->time), 0, sizeof(struct timeval));
+	tcp->state= TCP_CLOSED;
+
+	tcp->ack= 0;
+	tcp->win= sizeof(tcp->in.data) - 1;
+
+	return(SUCCESS);
+}
+
+
+
+/*
+ * Function: tcp_open()
+ *
+ * Performs an open (active or passive) on a TCP socket
+ */
+
+int tcp_open(struct iface_data *idata, struct tcp *tcb, unsigned int mode){
+	if(mode == OPEN_ACTIVE){
+		tcb->state= TCP_SYN_SENT;
+		tcb->flags= TH_SYN;
+		tcb->snd_una= random();
+		tcb->snd_nxt= tcb->snd_una + 1;
+		tcb->pending_write_f= TRUE;
+		return(SUCCESS);
+	}
+	else if(mode == OPEN_PASSIVE){
+		tcb->state= TCP_LISTEN;
+		return(SUCCESS);
+	}
+
+	return(FAILURE);
+}
+
+
+/*
+ * Function: tcp_close()
+ *
+ * Performs a close on a TCP socket
+ */
+
+int tcp_close(struct iface_data *idata, struct tcp *tcb){
+	tcb->fin_flag= TRUE;
+	tcb->fin_seq= tcb->snd_nxt;
+	tcb->pending_write_f= TRUE;
+	return(SUCCESS);
+}
+
+
+/*
+ * Function: tcp_send()
+ *
+ * Sends data over TCP (actually copies it to the TCP send buffer)
+ */
+
+int tcp_send(struct iface_data *idata, struct tcp *tcb, unsigned char *data, unsigned int nbytes){
+	if(tcb->fin_flag == TRUE)
+		return(-1);
+	else
+		return(queue_data( &(tcb->out), data, nbytes));
+}
+
+
+/*
+ * Function: tcp_receive()
+ *
+ * Receive data from a TCP socket
+ */
+
+int tcp_receive(struct iface_data *idata, struct tcp *tcb, unsigned char *data, unsigned int nbytes){
+	unsigned int	r;
+
+	r= dequeue_data(&(tcb->in), data, nbytes);
+
+	if(r > 0)
+		tcb->pending_write_f= TRUE;
+
+	return(r);
+}
+
+
+/*
+ * Function: tcp_input()
+ *
+ * Processes an incoming TCP segment
+ */
+
+int tcp_input(struct iface_data *idata, struct tcp *tcb, const u_char *pktdata, struct pcap_pkthdr *pkthdr, struct packet *packet){
+	return(SUCCESS);
+}
+
+
+
+/*
+ * Function: tcp_output()
+ *
+ * Sends TCP segments as necessary
+ */
+
+int tcp_output(struct iface_data *idata, struct tcp *tcb, struct packet *packet, struct timeval *curtime){
+	/* Placeholder */
+	return(SUCCESS);
+}
+
+
+
+/*
+ * Function: is_valid_tcp_segment()
+ *
+ * Performs sanity checks on an incomming TCP/IPv6 segment
+ */
+
+int is_valid_tcp_segment(struct iface_data *idata, const u_char *pktdata, struct pcap_pkthdr *pkthdr){
+	struct ether_header	*pkt_ether;
+	struct ip6_hdr		*pkt_ipv6;
+	struct tcp_hdr		*pkt_tcp;
+	unsigned char		*pkt_end;
+
+	pkt_ether = (struct ether_header *) pktdata;
+	pkt_ipv6 = (struct ip6_hdr *)((char *) pkt_ether + idata->linkhsize);
+	pkt_tcp = (struct tcp_hdr *) ((char *) pkt_ipv6 + MIN_IPV6_HLEN);
+
+	pkt_end = (unsigned char *) pktdata + pkthdr->caplen;
+
+	/* XXX: We are assuming no extension headers on incoming packets -- this should be improved! */
+
+	/* The packet length is the minimum of what we capured, and what is specified in the
+	   IPv6 Total Lenght field
+	 */
+	if( pkt_end > ((unsigned char *)pkt_tcp + pkt_ipv6->ip6_plen) )
+		pkt_end = (unsigned char *)pkt_tcp + pkt_ipv6->ip6_plen;
+
+	/*
+	   Discard the packet if it is not of the minimum size to contain a TCP header
+	 */
+	if( (pkt_end - (unsigned char *) pkt_tcp) < sizeof(struct tcp_hdr)){
+		return FALSE;
+	}
+
+	/* Check that the TCP checksum is correct */
+	if(in_chksum(pkt_ipv6, pkt_tcp, pkt_end-((unsigned char *)pkt_tcp), IPPROTO_TCP) != 0){
+		return FALSE;
+	}
+
+
+	/* XXX: Should perform additional checks on the IPv6 header */
+	/*
+	   Sanity checks on the Source Address and the Destination Address
+	 */
+	if(IN6_IS_ADDR_UNSPECIFIED(&(pkt_ipv6->ip6_src)) || IN6_IS_ADDR_UNSPECIFIED(&(pkt_ipv6->ip6_dst))){
+		return FALSE;
+	}
+
+	if(IN6_IS_ADDR_MULTICAST(&(pkt_ipv6->ip6_src)) || IN6_IS_ADDR_MULTICAST(&(pkt_ipv6->ip6_dst))){
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
